const createRollupConfig = require('../../src/config/rollup.config.default');
const path = require('path');
const rollup = require('rollup');
const jsdom = require('jsdom');

jest.setTimeout(9000000);
describe('the bundling process', () => {
  it('should obey minification flag', () =>
    Promise.all(
      [true, false].map(async minify => {
        const config = await createRollupConfig({
          dir: path.join(__dirname, 'minify'),
          minify
        });

        const bundle = await rollup.rollup(config);
        const {output} = await bundle.generate(config.output);
        const {code} = output['mashup.js'];
        const mangledVariableName = 'thisVariableNameShouldBeMangledByMinifier';
        if (minify) {
          expect(code.indexOf(mangledVariableName)).toBeLessThan(0);
        } else {
          expect(code.indexOf(mangledVariableName)).toBeGreaterThan(0);
        }
      })
    ));
  it('should remove all comments from the bundle, even if minify is false', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'require-in-comments'),
      minify: false
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);
    const {code} = output['mashup.js'];
    expect(code.indexOf('require("in-comment")')).toBeLessThan(0);
    expect(code.indexOf('require("in-string")')).toBeGreaterThan(0);
  });
  it('should produce a bundle that calls into mashups API with addMashup or addModule and correctly handles dependencies', () =>
    Promise.all(
      ['mashup'].map(async target => {
        const config = await createRollupConfig({
          dir: path.join(__dirname, 'dependency'),
          target,
          minify: true
        });

        const bundle = await rollup.rollup(config);
        const {output} = await bundle.generate(config.output);
        const {code} = output['mashup.js'];

        if (window.tau) {
          delete window.tau;
        }
        const notRemappedMock = jest.fn();
        const remappedMock = jest.fn();
        let moduleResult;

        // fake global will be called in actual mashup code to verify that it executed
        window.tau = {mashups: {}, fakeGlobal: jest.fn()};
        window.tau.mashups.addDependency = jest.fn(() => window.tau.mashups);
        window.tau.mashups.addMashup = jest.fn(initializer =>
          initializer(notRemappedMock, remappedMock)
        );
        window.tau.mashups.addModule = jest.fn((moduleName, initializer) => {
          moduleResult = initializer(notRemappedMock, remappedMock);
        });

        // jsdom here, so we can eval generated mashup code straight away
        eval(code);

        expect(window.tau.mashups.addDependency).toHaveBeenCalledWith(
          'not-remapped-dependency'
        );
        expect(window.tau.mashups.addDependency).toHaveBeenCalledWith(
          'remapped-dependency'
        );

        if (target === 'mashup') {
          expect(window.tau.mashups.addMashup).toHaveBeenCalled();
          // mashups should be executed right away
        } else {
          expect(window.tau.addModule).toHaveBeenCalled();

          // modules should not execute their code until their initialize is called
          expect(notRemappedMock).not.toHaveBeenCalled();
          expect(remappedMock).not.toHaveBeenCalled();
          expect(window.tau.fakeGlobal).not.toHaveBeenCalled();

          expect(typeof moduleResult.initialize).toBe('function');

          moduleResult.initialize();
        }

        expect(notRemappedMock).toHaveBeenCalled();
        expect(remappedMock).toHaveBeenCalled();
        expect(window.tau.fakeGlobal).toHaveBeenCalled();
      })
    ));

  it('should produce a bundle that uses CSS modules correctly and appends all used classes as style tag', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'css-modules'),
      target: 'mashup',
      minify: true
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);
    const {code} = output['mashup.js'];

    if (window.tau) {
      delete window.tau;
    }

    // fake global will be called in actual mashup code to verify that it executed
    window.tau = {mashups: {}, dumpStyles: jest.fn()};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer();

    // jsdom here, so we can eval generated mashup code straight away
    eval(code);

    expect(window.tau.dumpStyles).toHaveBeenCalledWith({
      ['test-class']: 'test-class__css-modules__1-0-0',
      foo: 'foo__css-modules__1-0-0',
      bar: 'bar__css-modules__1-0-0',
      imported: 'imported__css-modules__1-0-0',
      nested: 'nested__css-modules__1-0-0'
    });

    const styleTag = document.getElementsByTagName('style')[0];
    expect(
      styleTag.innerHTML.indexOf('.test-class__css-modules__1-0-0')
    ).toBeGreaterThanOrEqual(0);
    expect(styleTag.innerHTML.indexOf('.global')).toBeGreaterThanOrEqual(0);
    expect(
      styleTag.innerHTML.indexOf('.global__css-modules__1-0-0')
    ).toBeLessThan(0);
  });

  it('should throw when dynamic imports are used and basePath is not specified', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'code-splitting'),
      target: 'mashup',
      minify: true
    });

    const bundle = await rollup.rollup(config);
    await expect(bundle.generate(config.output)).rejects.toThrow(
      /basePath not specified/
    );
  });

  it('should perform code-splitting and load chunks when they are required', async () => {
    const chunksBasePath = 'http://fake-url';
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'code-splitting'),
      target: 'mashup',
      basePath: chunksBasePath,
      minify: true
    });

    const {mashupChunksGlobalName} = config.plugins.find(
      plugin => plugin.name === 'mashup-entry-plugin'
    ).settings;

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);
    const chunkNames = Object.keys(output);
    // + mashup.js (entry point) includes 2 common chunks - commonChunk.js
    // used by chunk1.js and chunk2.js and regenerator runtime used by
    // entry and arbitaryFolder/chunk
    // + chunk1.js
    // + chunk2.js
    // + chunk.js
    // + arbitaryFolder/chunk.js (name clashing is intentional)
    // + manifest.cfg (because target: 'mashup')
    expect(chunkNames).toHaveLength(6);

    // we can't use Jest's built-in jsdom because we need to override script loading options
    const chunkRequests = [];
    const dom = new jsdom.JSDOM('', {
      runScripts: 'dangerously',
      resources: 'usable',
      resources: new class extends jsdom.ResourceLoader {
        fetch(url) {
          if (url.startsWith(chunksBasePath)) {
            const chunkName = url.substring(chunksBasePath.length + 1);
            chunkRequests.push(chunkName);
            return Promise.resolve(Buffer.from(output[chunkName].code));
          }
        }
      }()
    });

    const {window} = dom;

    // fake global will be called in actual mashup code to verify that it executed
    window.tau = {mashups: {}, dumpStyles: jest.fn()};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer();
    const initChunkCode = output['mashup.js'].code;
    // we expect our entry point to also contain commons chunk - in this case, regenerator runtime
    expect(initChunkCode.indexOf('common chunk')).toBeGreaterThan(0);

    await new Promise(resolve => {
      window.tau.initChunksLoaded = resolve;
      window.eval(initChunkCode);
      // by the time entry chunk is executed on the page, we expect to have 3
      // chunks available: mashup.js (entry), common chunk with regenerator and
      // common chunk from commonChunk.js
      expect(Object.keys(window[mashupChunksGlobalName])).toHaveLength(3);
    });
    // when the above promise resolves, we can check loaded chunks
    // we expect all chunks except the last one to be present
    expect(Object.keys(window[mashupChunksGlobalName])).toHaveLength(6);
    await new Promise(resolve => {
      window.tau.lastChunkLoaded = resolve;
      window.tau.loadLastChunk();
    });

    // the last chunk is loaded only after arbitary lastChunkLoaded function call
    // there are 7 chunks total on the page - entry, 2 common and 4 from dynamic imports
    expect(Object.keys(window[mashupChunksGlobalName])).toHaveLength(7);
    // we expect dynamic imports to be reused once they are loaded, i.e.
    // there should be no more than 4 requests for dynamic chunks
    expect(chunkRequests).toHaveLength(4);
  });

  it('should support typescript and tsx out of the box', async () => {
    const config = await createRollupConfig({
      dir: path.join(__dirname, 'typescript'),
      target: 'mashup'
    });

    const bundle = await rollup.rollup(config);
    const {output} = await bundle.generate(config.output);
    const {code} = output['mashup.js'];

    if (window.tau) {
      delete window.tau;
    }
    const fakeComponentConstructor = jest.fn();
    const fakeReact = {
      createElement: jest.fn((type, props, children) => ({
        type,
        props,
        children
      })),
      Component: function(...args) {
        fakeComponentConstructor(...args);
      }
    };

    // fake global will be called in actual mashup code to verify that it executed
    window.tau = {mashups: {}};
    window.tau.mashups.addDependency = () => window.tau.mashups;
    window.tau.mashups.addMashup = initializer => initializer(fakeReact);

    eval(code);
    expect(fakeReact.createElement).toHaveBeenCalledWith('div', null, {
      type: window.tau.smallComponent,
      children: undefined,
      props: {id: 5}
    });
    expect(fakeComponentConstructor).toHaveBeenCalledWith({name: 'foo'});
    expect(window.tau.reactElement.type).toBe('div');
    expect(Object.getPrototypeOf(window.tau.reactElement.children.type)).toBe(
      fakeReact.Component
    );
  });
});
